-- concrete syntax
module Concrete where

import Util
import Abstract (Name,Co,Pos,Sized,Dec(..))

data Expr = Set
          -- size type
          | Size 
          | Succ Expr
          | Infty
          | Max
          -- 
          | App Expr [Expr]
          | Lam Name Expr
          | Case Expr [Clause]
          | LLet Name Expr Expr Expr -- local let
          | Pi Dec Name Expr Expr
          | Ident Name 
          | Unknown
          -- singleton type
          | Sing Expr Expr 
          deriving (Eq)

instance Show Expr where
    show = prettyExpr

data Declaration = DataDecl Name Sized Co Telescope Type [Constructor]
                 | FunDecl Co TypeSig [Clause] 
                 | LetDecl Bool TypeSig Expr 
                 | MutualDecl [Declaration] -- bool = if eval
                   deriving (Eq,Show)

data TypeSig = TypeSig Name Type
             deriving (Eq,Show)

type Type = Expr

type Constructor = TypeSig

data TBind = TB Dec Name Type 
    | PosTB Dec Name Type
      deriving(Eq,Show)

type Telescope = [TBind]

data Clause = Clause [Pattern] (Maybe Expr) -- Nothing for absurd pattern clause  
            deriving (Eq,Show)

data Pattern = ConP Name [Pattern] -- (c ps)
             | SuccP Pattern       -- ($ p)
             | DotP Expr           -- .p
             | IdentP Name         -- x
             | SizeP Name Name     -- (x > y) 
             | AbsurdP             -- ()
               deriving (Eq,Show)

type Case = (Pattern,Expr)

----



prettyExpr :: Expr -> String
prettyExpr e = 
    case e of
      Set             -> "Set"
      Size            -> "Size" 
      Max             -> "max" 
      Succ e          -> "($ " ++ prettyExpr e ++ ")"
      Infty           -> "#"
      App e1 el       -> "(" ++ prettyExprs (e1:el) ++ ")"
      Lam x e1        -> "(\\" ++ x ++ " -> " ++ prettyExpr e1 ++ ")"
      Case e cs       -> "case " ++ prettyExpr e ++ " { " ++ Util.showList "; " prettyCase cs ++ " } "
      LLet n t1 e1 e2 -> "(let " ++ n ++ " : " ++ prettyExpr t1 ++ " = " ++ prettyExpr e1 ++ " in " ++ prettyExpr e2 ++ ")" 
      Ident n         -> n
      Unknown         -> "_"
      Sing e t        -> "<" ++ prettyExpr e ++ " : " ++ prettyExpr t ++ ">"
      Pi dec x t1 t2  -> parens $ (if erased dec then brackets binding
                                    else if x=="" then s1 else parens binding) 
                                  ++ " -> " ++ prettyExpr t2
         where s1 = prettyExpr t1
               binding = if x == "" then s1 else x ++ " : " ++ s1 

prettyCase (Clause [p] Nothing)  = prettyPattern p 
prettyCase (Clause [p] (Just e)) = prettyPattern p ++ " -> " ++ prettyExpr e

prettyPattern :: Pattern -> String
prettyPattern (ConP c ps) = parens $ foldl (\ acc p -> acc ++ " " ++ prettyPattern p) c ps
prettyPattern (SuccP p) = parens $ "$ " ++ prettyPattern p
prettyPattern (DotP e)  = "." ++ prettyExpr e
prettyPattern (IdentP x) = x
prettyPattern (SizeP x y) = parens $ x ++ " > " ++ y
prettyPattern (AbsurdP) = parens ""

prettyExprs :: [Expr] -> String
prettyExprs = Util.showList " " prettyExpr

teleToType :: Telescope -> Type -> Type
teleToType [] t = t
teleToType (TB dec n t:tel) t2 = Pi dec n t (teleToType tel t2)
teleToType (PosTB dec n t:tel) t2 = Pi dec n t (teleToType tel t2)
