{ 
module Parser where

import qualified Lexer as T
import qualified Concrete as C

import Abstract (Dec(..))
import qualified Abstract as A
import Abstract (Name)
}

%name parse
%tokentype { T.Token }
%error { parseError }

%token

id      { T.Id $$ _ }
number  { T.Number $$ _ }
data    { T.Data _ }
codata  { T.CoData _ }
sized   { T.Sized _ }
mutual  { T.Mutual _ }
fun     { T.Fun _ }
cofun   { T.CoFun _ } 
case    { T.Case _ }
let     { T.Let _ }
in      { T.In _ }
eval    { T.Eval _ }
set     { T.Set _ }
size    { T.Size _ }
infty   { T.Infty _ }
succ    { T.Succ _ }
max     { T.Max _ }
'<'     { T.AngleOpen _ }
'>'     { T.AngleClose _ }
'{'     { T.BrOpen _ }
'}'     { T.BrClose _ }
'['     { T.BracketOpen _ }
']'     { T.BracketClose _ }
'('     { T.PrOpen _ }
')'     { T.PrClose _ }
'|'     { T.Bar _ }
';'     { T.Sem _ }
':'     { T.Col _ }
'.'     { T.Dot _ }
'->'    { T.Arrow _ }
'='     { T.Eq _ }
'+'     { T.Plus _ }
'*'     { T.Times _ } -- UNUSED
'\\'    { T.Lam _ }
'_'     { T.Underscore _ }

%%

TopLevel :: { [C.Declaration] }
TopLevel : Declarations { reverse $1}


Declarations :: { [C.Declaration] }
Declarations : {- empty -} { [] }
             | Declarations Declaration { $2 : $1 }

Declaration :: { C.Declaration }
Declaration : Data { $1 }
           | CoData { $1 }
           | SizedData { $1 }
           | SizedCoData { $1 }
           | Fun { $1 }
           | CoFun { $1 }
           | Mutual { $1 }
           | Let { $1 }

Data :: { C.Declaration }
Data : data Id DataTelescope ':' Expr '{' Constructors '}' 
   { C.DataDecl $2 A.NotSized A.Ind $3 $5 (reverse $7) }

SizedData :: { C.Declaration }
SizedData : sized data Id DataTelescope ':' Expr '{' Constructors '}' 
   { C.DataDecl $3 A.Sized A.Ind $4 $6 (reverse $8) }

CoData :: { C.Declaration }
CoData : codata Id DataTelescope ':' Expr '{' Constructors '}' 
       { C.DataDecl $2 A.NotSized A.CoInd $3 $5 (reverse $7) }

SizedCoData :: { C.Declaration }
SizedCoData : sized codata Id DataTelescope ':' Expr '{' Constructors '}' 
       { C.DataDecl $3 A.Sized A.CoInd $4 $6 (reverse $8) }

Fun :: { C.Declaration }
Fun : fun TypeSig '{' Clauses '}' { C.FunDecl A.Ind $2 $4 }

CoFun :: { C.Declaration }
CoFun : cofun TypeSig '{' Clauses '}' { C.FunDecl A.CoInd $2 $4  }



Mutual :: { C.Declaration }
Mutual : mutual '{' Declarations '}' { C.MutualDecl (reverse $3) }
     

 
Let :: { C.Declaration }
Let : let TypeSig '=' Expr { C.LetDecl False $2 $4 } 
      | eval let TypeSig '=' Expr { C.LetDecl True $3 $5 } 

-----

Id :: { Name }
Id : number { $1 }
   | id { $1 }

Expr :: { C.Expr }
Expr : 
        TBind '->' Expr      { let (dec,n,t) = $1 in C.Pi dec n t $3 } 
       | '\\' Id '->' Expr                { C.Lam $2 $4 }
       | let Id ':' Expr '=' Expr in Expr { C.LLet $2 $4 $6 $8}
       | case Expr '{' Cases '}'          { C.Case $2 $4 }  
       | Expr1 '->' Expr                  { C.Pi (Dec False) "" $1 $3}
       | '[' Expr1 ']' '->' Expr          { C.Pi (Dec True) "" $2 $5}
       | Expr1                            { $1 }

Expr1 :: { C.Expr }
Expr1 : Expr2 { let list = reverse $1 in
               if length list == 1 then head list else C.App (head list) (tail list) 
	      }

-- gather applications
Expr2 :: { [C.Expr] }
Expr2 : Expr3 { [$1] }
       | Expr2 Expr3 { $2 : $1 }
--       | succ SE { [C.Succ $2] }

Expr3 :: { C.Expr }
Expr3 :  set                   { C.Set}
      | size                   { C.Size }
      | max                    { C.Max }
      | infty                  { C.Infty }
      | Id                     { C.Ident $1}
      | '<' Expr ':' Expr '>'  { C.Sing $2 $4 }
      | '(' Expr ')'           { $2 }
      | '_'                    { C.Unknown }
      | succ Expr3             { C.Succ $2 }  -- succ is a prefix op

{- RETIRED
SE :: { C.Expr}
SE : succ SE { C.Succ $2}
SE : Expr3 { $1}
  -}

TypeSig :: { C.TypeSig }
TypeSig : Id ':' Expr { C.TypeSig $1 $3 }

Constructor :: { C.Constructor }
Constructor : TypeSig { $1 }

Constructors :: { [C.Constructor ] }
Constructors :
      Constructors ';' Constructor { $3 : $1 }
    | Constructors ';' { $1 }
    | Constructor { [$1] }
    | {- empty -} { [] }

Cases :: { [C.Clause] }
Cases : Pattern '->' Expr ';' Cases  { (C.Clause [$1] (Just $3)) : $5 }
      | Pattern '->' Expr            { (C.Clause [$1] (Just $3)) : [] }
      | Pattern ';' Cases            { (C.Clause [$1] Nothing) : $3 }
      | Pattern                      { (C.Clause [$1] Nothing) : [] }
      | {- empty -}                  { [] }
      
Clause :: { C.Clause }
Clause : Id LHS '=' Expr { C.Clause $2 (Just $4) }
       | Id LHS          { C.Clause $2 Nothing }

LHS :: { [C.Pattern] }
LHS : Patterns { reverse $1 }

Patterns :: { [C.Pattern] }
Patterns : {- empty -} { [] }
    | Patterns Pattern { $2 : $1 }

Pattern :: { C.Pattern }
Pattern : ConP               { $1            }
        | Id                 { C.IdentP $1   }
        | '.' Expr3          { C.DotP $2     }
        | '(' Id '>' Id ')'  { C.SizeP $2 $4 } 
--        | '*'       { C.AbsurdP }

ConP :: { C.Pattern }
ConP : '(' Id Patterns ')' { C.ConP $2 (reverse $3) }
     | '(' succ SP ')'     { C.SuccP $3             }
     | '(' ')'             { C.AbsurdP              }

SP :: { C.Pattern}
SP : succ SP {C.SuccP $2}
   | Pattern { $1 }

Clauses :: { [C.Clause] }
Clauses : RClauses { reverse $1 }

RClauses :: { [C.Clause ] }
RClauses :
   RClauses ';' Clause { $3 : $1 }
 | RClauses ';' { $1 }
 | Clause { [$1] }
 | {- empty -} { [] }


TBind :: { (Dec,Name,C.Expr) }
TBind :  '(' Id ':' Expr ')' { (Dec False,$2,$4) } -- ordinary binding
      |  '[' Id ':' Expr ']' { (Dec True,$2,$4) }  -- erased binding

TBindSP :: { C.TBind }
TBindSP :  '(' '+' Id ':' Expr ')' { C.PosTB (Dec False) $3 $5 }
        |  '[' '+' Id ':' Expr ']' { C.PosTB (Dec True) $3 $5 }

TBindNP :: { C.TBind }
TBindNP :  '(' Id ':' Expr ')' { C.TB (Dec False) $2 $4 }
        |  '[' Id ':' Expr ']' { C.TB (Dec True) $2 $4 }

DataTelescope :: { C.Telescope }
DataTelescope :  {- empty -} { [] }
                | TBindSP DataTelescope { $1 : $2 } 
                | TBindNP DataTelescope { $1 : $2 } 

{

parseError :: [T.Token] -> a
parseError [] = error "Parse error at EOF"
parseError (x : xs) = error ("Parse error at token " ++ T.prettyTok x) 

}

 



